/***********************************************************************
 * interface.c: rudimentary command-line interface to popmail
 * 				interactive mode.
 * 	$Revision: 1.14 $
 *
 * 	Copyright (c) 2003, 2004 L. Alberto Gimnez Benito	<algibe@teleline.es>
 * 
 *  This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * **********************************************************************/


#include "interface.h"
#include "filter.h"

#define _(string) gettext (string)

#define MAX_LINE 256

static const char rcsid[]="$Id: interface.c,v 1.14 2004/09/10 19:47:48 alberto Exp $";
static int working_server = 0; /* we work on this server */

static int have_mail (int num_servers);
static char parse_command (char *line, int *arg);
static int interactive_help (void);
static int print_headers (int message);
static int print_all_headers (int message);
static int change_working_server (int arg, int num_servers);
static int delete_message (int message);
static int undelete_message (int message);
static int print_servers (int num_servers);
static int filter_mode (void);
static void parse_deletion(char *line);
static action_t ask_action(void);

	
int iface_init (int num_servers, char *filterfile)
{
	char line[MAX_LINE];
	char command; /* command */
	int done = 0;
	int arg = -1;	/* argument */
	/*int i, j;*/
	int nfilters; /* number of filters before entering interactive mode */
	
	printf (_("\nNow you are in interactive mode. enter 'h' command to"
				 " view interactive mode help, or 'q' to quit.\n\n"
				 "Available servers: \n"));
	
	print_servers (num_servers);

	/* store original #filters to decide if new filters are added */
	nfilters = num_filters (&filters);

	do {
		printf ("> ");
		fgets (line, MAX_LINE, stdin);
		command = parse_command (line, &arg);
		
		if (server_msgs (servidores[working_server]) == 0 && 
				command != 'c' && command != 'q' && command != 's' &&
				command != 'h') { /* no mail */
			printf (_("No mail in the server %s. Please change working server.\n"),
				server_name (servidores[working_server]));
			continue;
		}
		
		if (arg == -1) { /* invalid argument */
			printf (_("Invalid argument\n"));
			continue;
		}

					
		switch (command) {
			case 'h': 
				interactive_help ();
				break;
						 
			case 'q':
				done = 1;
				break;

			case 'l': /* list 'From' and 'Subject' headers of message #ARG */
				print_headers (arg);
				break;

			case 'a': /* list all headers of message #arg */
				print_all_headers (arg);
				break;
						 
			case 'c': /* change to server #ARG */
				change_working_server (arg, num_servers);
				break;

			case 'd': /* delete message number ARG */
				parse_deletion(line);
			//	delete_message (arg);
				break;

			case 'u':
				undelete_message (arg);
				break;

			case 's': /* show server names */
				print_servers (num_servers);
				break;			

			case 'f': /* filter mode */
				filter_mode ();
				break;

			default : /* unknown command */
				printf (_("%c: unknown command\n"), command);
				break;
			
		}
		
	} while (!done);

	/*
	if (opt_dummy)
		printf (_("Dummy mode: messages will not be deleted\n"));
	else
		real_deletion (num_servers);
	*/
	
	/*--- Free memory allocated for the message list: --------------*/
	/*for (j = 0; j < num_servers; j++) {
    	for (i=0; i<servidores[j]->num_msgs; i++ )
        	 msg_free (servidores[j]->msgs[i]);
      free (servidores[j]->msgs);
	}*/
      /*--------------------------------------------------------------*/

	if (num_filters (&filters) > nfilters) /* filters were added */
		WriteFilterFile (filterfile, &filters, nfilters); /* write them to disk */

	return (have_mail (num_servers) ? 0 : 1); /* return 0 if mail */
}

/* have_mail: returns 1 if there is mail remaining on any server, and 0
 * otherwise */
int have_mail (num_servers)
{
	int i;

	for (i = 0; i < num_servers; i++) {
		if (server_kept (servidores[i]))
			break; /* continue until a server has messages or all servers checked */
	}
	
	return (i != num_servers || server_kept (servidores[num_servers-1]));
}
	
/* get_command: gets a command in line (first character) or NULL on
 * invalid first character (not alphabetical) */
char get_command (char *line)
{
	for (; !isspace (*line); line++)
		; /* look for non-blank char */
	
	line--;
	if (!isalpha (*line)) /* no character */
		return 0;	
	else
		return (*line);
}

/* get_arg: returns the integer value of a command argument, 
 * 0 if no argument and -1 if invalid argument */
int get_arg (char *line)
{
	int arg; /* number returned */
	char number[5];

	GetnWord (line, 2, number, sizeof (number)); /* shoud be the arg */
	if (number[0] == '\0') /* no argument */
		arg = 0;
	else if (!isdigit (number[0])) /* invalid argument */
		arg = -1;
	else { /* is a digit. Must be positive and non-zero */
		arg = atoi (number);
		if (arg < 1)
			arg = -1; /* no '0' or negative arg allowed */
	}

	return arg;
}
	
/* parse_command: parses line, which contain a character (command) and
 * optionally an integer argument, arg. */
char parse_command (char *line, int *arg)
{
	char command;
	
	command = get_command (line);

	switch (command) { /* parse also the argument, if needed */
		case 'l':
		case 'a':
		case 'd':
		case 'u':
		case 'c':
		case 'f':
			*arg = get_arg (line);
			break;
			
		default : /* command has no arguments */
			*arg = 0;
			break;
	}

	return command;
}

/* prints interactive mode help */
int interactive_help (void)
{
	char *ver = VERSION;
	printf (_("\nPOPMail version %s interactive mode.\n"
			"Copyright (C) 2003 L. Alberto Gimnez.\n"
			"Allowed commands and their (optional or not) arguments are: \n"
			"h -- Prints this help\n"
			"l -- List 'From' and 'Subject' headers of all messages in current server\n"
			"\tOptional argument: an integer value. Only headers of this\n"
			"\tmessage will be shown.\n"
			"a -- List all headers of all messages in the current server\n"
			"\tOptional argument: an integer value. Only heaaders of this\n"
			"\tmessage will be shown.\n"
			"d -- Delete messages in the current server\n"
			"     Ranges (1-10) and multiple args are allowed (2 4 6-9)\n"
			"f -- Enter filtering mode\n"
			"u number -- Undelete message number NUMBER in the current server\n"
			"c number -- Changes current server to NUMBER\n"
			"s -- Show all server names. Working server is marked by a '->'\n"
			"q -- Exit interactive mode\n\n"), ver);

	return 0;
}

/* print_headers: prints From and Subject headers of msg #message, or
 * all the messages headers if message == 0 */
int print_headers (int message)
{
	if (server_kept (servidores[working_server]) == 0) { /* no mail */
		printf (_("No mail remaining in the server %s\n"),
				server_name (servidores[working_server]));
		return 1;
	}
	if (message == 0) { /* no argument->print all messages */
		for (message = 0; message < server_kept (servidores[working_server]); 
			message++) {
			if (!msg_deleted (servidores[working_server]->msgs[message]))
				printf(_("Msg #%d: From: %s\n"
							"        Subject: %s\n\n"), message+1, 
					msg_from (servidores[working_server]->msgs[message]),
					msg_subject (servidores[working_server]->msgs[message]));
		}
	}
	else { /* message != 0 */
		if (message > server_msgs (servidores[working_server])) { /* out of range */
			printf (_("No message #%d in the server %s\n"), message, 
					server_name (servidores[working_server]));
		}
		else { /* not out of range */
			if (msg_deleted (servidores[working_server]->msgs[message-1]))
				printf(_("Warning: message marked for deletion\n"));

			printf(_("Msg #%d: From: %s\n"
						"        Subject: %s\n\n"), message,
					msg_from (servidores[working_server]->msgs[message-1]),
					msg_subject (servidores[working_server]->msgs[message-1]));
		} /* not out of range */
	} /* message == 0 */
	
	return 0;
}

/* print_all_headers: prints all headers of msg #message, or all the
 * headers of all the messages if message == 0 */
int print_all_headers (int message)
{
	if (message == 0) { /* no argument->print all messages */
		if (server_kept (servidores[working_server]) == 0) /* no mail */
			printf (_("No mail remaining in the server %s\n"),
					server_name (servidores[working_server]));
		else { /* there is mail */
			for (message = 0; message < server_kept (servidores[working_server]); 
				message++) {
				if (!msg_deleted (servidores[working_server]->msgs[message]))
					printf(_("======== Msg #%d ========\n%s\n\n"), message+1,
							msg_header (servidores[working_server]->msgs[message]));
			}
		}
	}
	else { /* message != 0 */
		if (message > server_msgs (servidores[working_server])) { /* out of range */
			printf (_("No message #%d in the server %s\n"), message, 
					server_name (servidores[working_server]));
		}
		else { /* not out of range */
			if (msg_deleted (servidores[working_server]->msgs[message-1]))
				printf(_("Warning: message marked for deletion\n"));

			printf(_("======== Msg #%d ========\n%s\n\n"), message,
					msg_header (servidores[working_server]->msgs[message-1]));
		} /* not out of range */
	} /* message == 0 */
			
	return 0;
}

/* change_working_server: changes current server where we are working */
int change_working_server (int arg, int num_servers)
{
	if (arg == 0)  /* no argument */
		printf (_("Change server command needs an argument\n"));
	else if (arg == working_server+1) /* already on that server */
		printf (_("You are already working at server %d\n"), arg);
	else if (arg > num_servers) /* out of range */
		printf (_("No server #%d\n"), arg);
	else { /* make real change */
		working_server = arg-1;
		printf (_("Working server switched to #%d (%s)\n"),
				working_server+1, server_name (servidores[working_server]));
	}
		
	return 0;
}

/* delete_message: marks as deleted message #MESSAGE at WORKING_SERVER */
int delete_message (int message)
{ 
	if (message == 0) /* no argument supplied */
		printf (_("Delete command needs an argument\n"));
	else if (message > server_msgs (servidores[working_server])) /* out of range */
		printf (_("There is no message #%d in the server %s\n"),
				message, server_name (servidores[working_server]));
	else { /* delete message */
		if (!msg_deleted (servidores[working_server]->msgs[message-1]))
			msg_set_deleted (servidores[working_server]->msgs[message-1]);
			/* don't want to delete really, so we can undelete later */
		else /* marked yet as deleted */
			printf (_("Message #%d is already marked for deletion\n"), message);
	}
	
	return 0;
}

/* undelete_message: clears deletion mark to MESSAGE at WORKING_SERVER */
int undelete_message (int message)
{
	if (message == 0) /* no argument supplied */
		printf (_("Undelete command needs an argument\n"));
	else if (message > server_msgs (servidores[working_server])) /* out of range */
		printf (_("There is no message #%d in the server %s\n"),
				message, server_name (servidores[working_server]));
	else { /* undelete message */
		if (!msg_deleted (servidores[working_server]->msgs[message-1]))
			printf (_("Message #%d is not marked for deletion\n"), message);
		else /* marked as deleted */
			msg_clear_deleted (servidores[working_server]->msgs[message-1]);
	}

	return 0;
}
/* print_servers: prints all configured servers via popmailrc */
int print_servers (int num_servers)
{
	int i;
	
	for (i = 0; i < num_servers; i++) {
		if (i == working_server) /* put the active mark */
			printf(_("\t%d: -> "), i+1);
		else
			printf(_("\t%d:    "), i+1);

		printf ("%s@%s\n", server_user (servidores[i]), server_name (servidores[i]));
	}
	return 0;
}


/* shows help for interactive filter mode */
void filter_help (void)
{
	char *ver = VERSION;
	char *docdir =	 DESTDOCDIR;

	printf (_("POPMail version %s interactive filter mode.\n"
				"Copyright (C) 2003 L. Alberto Gimnez\n\n"
				"From this mode you can add dinamically filters to your\n"
				"filterfile's rules, from messages kept in the server.\n\n"
				"You can also add manual filters, but you must follow the\n"
				"rules indicated in filterfile.example (should be installed\n"
				"inside %s).\n"
				"These are available commands: \n\n"
			"Allowed commands and their (optional or not) arguments are: \n"
			"h -- Prints this help\n"
			"l -- List 'From' and 'Subject' headers of all messages in current server\n"
			"\tOptional argument: an integer value. Only headers of this\n"
			"\tmessage will be shown.\n"
			"a -- List all headers of all messages in the current server\n"
			"\tOptional argument: an integer value. Only heaaders of this\n"
			"\tmessage will be shown.\n"
			"f number -- Adds a rule to full From header of message #NUMBER\n"
			"\tex: user@domain.com\n"
			"u number -- Adds a rule to sender username of message #NUMBER\n"
			"\tex: user@.*\n"
			"d number -- Adds a rule to sender domain of message #NUMBER\n"
			"\tex: .*@domain.com\n"
			"s number -- Adds a rule to Subject header of message #NUMBER\n"
			"m -- Manual filter: adds a string as a new filter. You\n"
			"\tshould avoid blanks (use '.*' instead)\n"
		/*	"p -- Applies filters to messages kept in the server\n"*/
			"v -- Shows all filter rules\n"
			"q -- Exit interactive mode\n\n"), ver, docdir);

}


/* parse_f_command: parses line, which contain a character (command) and
 * optionally an integer argument, arg. */
char parse_f_command (char *line, int *arg)
{
	char command;
	
	command = get_command (line);

	switch (command) { /* parse also the argument, if needed */
		case 'l':
		case 'a':
		case 'd':
		case 'u':
		case 'f':
		case 's':
			*arg = get_arg (line);
			break;
			
		default : /* command has no arguments */
			*arg = 0;
			break;
	}

	return command;
}


/* Enters filter mode: you can add filters to the filterfile from
 * here */
int filter_mode (void)
{
	int done = 0;
	char line[MAX_LINE];
	char command;
	int arg;
	/*int last;*/ /* last filterfile defined filter */
	
	printf (_("Now you are in filter mode. You are working in the account\n"
				"%s@%s. Enter 'h' command to get filter mode help.\n"),
			server_user (servidores[working_server]),
			server_name (servidores[working_server]));

	do {
		printf ("filter> ");
		fgets (line, MAX_LINE, stdin);
		command = parse_f_command (line, &arg);

#ifdef DEBUG
		printf ("Command: %c; arg: %i\n", command, arg);
#endif
	/*	
		if (server_msgs (servidores[working_server]) == 0 && 
				command != 'c' && command != 'q' && command != 's' &&
				command != 'h') {  no mail 
			printf (_("No mail in the server %s. Please change working server.\n"),
				server_name (servidores[working_server]));
			continue;
		}*/
			
		if (arg == -1) { /* invalid argument */
			printf (_("Invalid argument\n"));
			continue;
		}
	
		switch (command) {
			case 'h': 
				filter_help ();
				break;
							 
			case 'q':
				done = 1;
				break;
	
			case 'l': /* list 'From' and 'Subject' headers of message #ARG */
				print_headers (arg);
				break;
	
			case 'a': /* list all headers of message #arg */
				print_all_headers (arg);
				break;
						 
			case 's': /* filter subject header of msg #arg */
				add_filter_subject (msg_subject (servidores[working_server]->msgs[arg-1]),
						&filters, ask_action());
				break;
	
			case 'd': /* filter domain for msg #arg-1 */
				add_filter_domain (msg_from (servidores[working_server]->msgs[arg-1]),
						&filters, ask_action());
				break;
	
			case 'u': /* filter username for msg #arg-1 */
				add_filter_sender (msg_from (servidores[working_server]->msgs[arg-1]),
							&filters, ask_action());
				break;
	
			case 'f': /* filter full from header of msg #arg-1 */
				add_filter_from (msg_from (servidores[working_server]->msgs[arg-1]),
							&filters, ask_action());
				break;

/*			case 'p':*/ /* apply filters to msgs in the server */
				/* not implemented */
/*				break;*/

			case 'm': /* manual filter to a given string */
				printf (_("Enter regexp to filter: "));
				fgets (line, MAX_LINE, stdin);
				line[strlen(line)-1] = '\0'; /* skip \n */
				add_filter (line, &filters, ask_action());
				break;

			case 'v': /* view filters */
				print_filters (&filters);
				break;
	
	
			default : /* unknown command */
				printf (_("%c: unknown command\n"), command);
				break;
				
		}

	} while (!done);

	
	return 0;
}


/* 
 * parses LINE looking for message number ranges "n-nn" or message
 * numbers "n" and perform the deletion
 */
void parse_deletion(char line[])
{
	int nargs = count_words(line)-1; 
	int i;

	for (i = 1; i <= nargs; i++) { /*skip 'd' word */
		char arg[MAX_LINE];
		GetnWord(line, i+1, arg, sizeof(arg));

		/* check for ranges */
		if (strchr(arg, '-')) {
			char *first;
			char *last;
			int first_msg;
			int last_msg;

			first = strtok(arg, "-");
			first_msg = atoi(first);
			
			last = strtok(NULL, "-");
			last_msg = atoi(last);
#ifdef DEBUG
			printf("first: '%s' (%d)\nlast: '%s'(%d)\n\n", first,
					first_msg, last, last_msg);
#endif
			if (first_msg <= last_msg) {
				for (;first_msg <= last_msg; first_msg++) {
					delete_message(first_msg);
				}
			} else {
				printf(_("Ranges must be in ascending order.\n"));
			}
		} else { /* no range. Just delete one message */
			int msg = atoi(arg);
#ifdef DEBUG
			printf("msg: '%s' (%d)\n\n", arg, msg);
#endif
			delete_message(msg);
		}				
	}
}


/* Asks the user for the type of the filter */
action_t ask_action(void)
{
	char line[MAX_LINE];

	printf(_("Type of the filter?\n"
			"1. DENY\n"
			"2. ACCEPT\n"
			"type: "));

	do {
		printf(_("type: "));
		fgets(line, MAX_LINE, stdin);
	} while ((line[0] != '1') && (line[0] != '2'));
	
	return ((line[0] == '2') ? ACCEPT : DENY);
}
